/*
 * @lasPlot.java Version 1.1 4/25/2017
 *
 * Copyright (c) 2017 Kansas Geological Survey
 * 1930 Constant Avenue, Lawrence, Kansas, 66047, U.S.A.
 * All Rights Reserved.
 */

package las.plot;

import java.awt.*;
import java.awt.image.*;
import java.awt.geom.*;
import java.awt.Graphics2D.*;
import java.awt.event.*;

import plot.plotXYStruct;          //XY Plot Data Structure
import plot.plotSymbolStruct;
import plot.plotSymbol;

/** Class lasPlot
 *  <p> This class will create a  X-Y Plot for an array of
 *      the first 2 Principal Components vector data
 *
 *  @version 1.1 4/25/2017
 *  @author  John R. Victorine
 */

public class lasPlot extends Canvas
{
  private int              iPlot   = -1;
  private plotXYStruct     stPlot  = null;

  // -- Depth Ranges

  private double dStart  = 0.0;
  private double dEnd    = 0.0;

  // -- Borehole reference

  private double dRadius = 0.0;

  private int    iXo     = 0;
  private int    iYo     = 0;

  private int    iCircle = 360;
  private int    iXb[]   = null;
  private int    iYb[]   = null;

  // Plot Dimensions

  private int              iXBegin = 80;  // Left Edge of Plot Area
  private int              iYBegin = 60;  // Top Edge of Plot Area
  private int              iWidth  = 550; // Width of Plot Area
  private int              iHeight = 550; // Height of Plot Area

  // Plot Data

  private int              iRows = 0;  // Total Number of Rows
  private int              iX[];       // Array of X Data converted to Pixels
  private int              iY[];       // Array of Y Data converted to Pixels
  private plotSymbolStruct stSymbol[]; // Array of Symbol and Color for Point

  // Legend

  private int              iLegend = 0; // Total Number of Legends
  private plotSymbolStruct stLegend[];  // Array of Symbol and Color for Legend

  // Plot Symbols

  private plotSymbol pSymbol = new plotSymbol();

  /** Constructor lasPlot()
   *  <p> This is the Constructor for this class.
   *  @param iPlot  = the Plot Type
   *  @param stPlot = Plot Data Structure.
   */

  public lasPlot(int iPlot, plotXYStruct stPlot)
  {
    setBackground(Color.white);

    this.iPlot  = iPlot;
    this.stPlot = stPlot;

    if (stPlot.iWidth != 0)
      iWidth = stPlot.iWidth;

    if (stPlot.iHeight != 0)
      iHeight = stPlot.iHeight;

    iRows = stPlot.iRows;

    if (stPlot.stSymbols != null)
      this.stSymbol = stPlot.stSymbols;

    this.iLegend = stPlot.iLegend;
    if (stPlot.stLegend != null)
      this.stLegend = stPlot.stLegend;

    // Convert Raw Data to Pixels

//    getRedLineLimits();
    convertDataToPixels();
  }

  /* =============================================================== *
   * ----------------------- DELETE METHODS ------------------------ *
   * =============================================================== */

  /** Method close()
   * <p> This method will set strings and arrays to null so Java can
   *     automatically clear memory
   */

  public void close()
  {
    stPlot   = null;

    iXb      = null;
    iYb      = null;

    iX       = null; // Array of X Data converted to Pixels
    iY       = null; // Array of Y Data converted to Pixels
    stSymbol = null; // Array of Symbol and Color for Point
    stLegend = null; // Array of Symbol and Color for Legend

    if (pSymbol != null)
      pSymbol.close();
    pSymbol  = null;
  }

  /* ================================================================== *
   * ------------------------ SUPPORT METHODS ------------------------- *
   * ================================================================== */

  /** Method void convertDataToPixels()
   *  <p> This method will set the Plot Data
   */

  private void convertDataToPixels()
  {
    int    i      = 0;
    int    iRatio = 0;
    double dXMin  = 0;
    double dXMax  = 0;
    double dYMin  = 0;
    double dYMax  = 0;

    iRows = stPlot.iRows;
    iX    = new int [iRows];
    iY    = new int [iRows];

    // Set the X-Axis Data

    // IF the X-Axis Data is LINEAR THEN
    if (stPlot.iXLog == plot.plotXYStruct.LINEAR)
    {
      for (i=0; i<iRows; i++)
      {
        iRatio = (int)((double)iWidth *
                       (stPlot.dX[i] - stPlot.dXMinimum) /
                       (stPlot.dXMaximum - stPlot.dXMinimum));
        iX[i] = iXBegin + iRatio;
      }
    }
    // IF the X-Axis Data is LOG THEN
    else
    {
      dXMin = Math.log(stPlot.dXMinimum) / Math.log(10.0);
      dXMax = Math.log(stPlot.dXMaximum) / Math.log(10.0);

      for (i=0; i<iRows; i++)
      {
        iRatio = (int)((double)iWidth *
                       ((Math.log(stPlot.dX[i])/Math.log(10.0))-dXMin) /
                       (dXMax - dXMin));
        iX[i] = iXBegin + iRatio;
      }
    }

    // Set the Y-Axis Data

    // IF the Y-Axis Data is LINEAR THEN
    if (stPlot.iYLog == plot.plotXYStruct.LINEAR)
    {
      for (i=0; i<iRows; i++)
      {
        iRatio = (int)((double)iHeight *
                       (stPlot.dY[i] - stPlot.dYMinimum) /
                       (stPlot.dYMaximum - stPlot.dYMinimum));
        iY[i] = iHeight + iYBegin - iRatio;
      }
    }
    // IF the Y-Axis Data is LOG THEN
    else
    {
      dYMin = Math.log(stPlot.dYMinimum) / Math.log(10.0);
      dYMax = Math.log(stPlot.dYMaximum) / Math.log(10.0);

      for (i=0; i<iRows; i++)
      {
        iRatio = (int)((double)iHeight *
                       ((Math.log(stPlot.dY[i])/Math.log(10.0))-dYMin) /
                       (dYMax - dYMin));
        iY[i] = iHeight + iYBegin - iRatio;
      }
    }
  }

  /* ================================================================== *
   * -------------------------- GET METHODS --------------------------- *
   * ================================================================== */

  /** Method getPlotWidth()
   * <p> This method will return the Plot Width
   * @return iWidth = The Plot Width
   */

  public int getPlotWidth()  { return (iXBegin+iWidth+200); }

  /** Method getPlotHeight()
   * <p> This method will return the Plot Height
   * @return iHeight = The Plot Height
   */

  public int getPlotHeight() { return (iYBegin+iHeight); }

  /** Method getImage()
   * <p> This method will convert this plot to a buffered image
   * @return image = profile plot buffered image
   */

  public BufferedImage getImage()
  {
    BufferedImage image = null;
    int w = this.getWidth();
    int h = this.getHeight();

    image = new BufferedImage( w, h, BufferedImage.TYPE_INT_RGB );

    Graphics2D g2 = image.createGraphics();
    g2.setColor( Color.white );
    g2.fillRect( 0, 0, w, h );
    this.paint(g2);
    g2.dispose();

    return (image);
  }

  /* ================================================================== *
   * -------------------------- SET METHODS --------------------------- *
   * ================================================================== */

  /** Method setTitle1()
   * <p> This method will set the first user title
   * @param sText = the first title
   */

  public void setTitle1(String sText)
  {
    this.stPlot.sTitle1 = new String(sText);
    repaint();
  }

  /** Method setTitle2()
   * <p> This method will set the second user title
   * @param sText = the second title
   */

  public void setTitle2(String sText)
  {
    this.stPlot.sTitle2 = new String(sText);
    repaint();
  }

  /** Method setTitle3()
   * <p> This method will set the third user title
   * @param sText = the third title
   */

  public void setTitle3(String sText)
  {
    this.stPlot.sTitle3 = new String(sText);
    repaint();
  }

  /** Method setAxes()
   *  <p> This method will set up the limits for the X & Y Axes
   *  @param dMax = Maxium radius
   *  @param dInc = Increment value on the axis
   */

  public void setAxes(double dMax, double dInc)
  {
	double dMin = -1.0 * dMax;

    this.stPlot.dXMaximum   = dMax;
    this.stPlot.dXMinimum   = dMin;
    this.stPlot.dXIncrement = dInc;

    this.stPlot.dYMaximum   = dMax;
    this.stPlot.dYMinimum   = dMin;
    this.stPlot.dYIncrement = dInc;

    convertDataToPixels();
    repaint();
  }

  /** Method setXAxis()
   * <p> This method will reset the X-Axis Parameters
   * @param sText = The Axis Label
   * @param iGrid = The type of grid to create
   * @param dMax  = The Maximum axis value
   * @param dMin  = The Minimum axis value
   * @param dInc  = The Increment value on the axis - Linear Grid only
   * @param iCycle = The Number of Cycles - Log Grid only
   */

  public void setXAxis(String sText,
                       int    iGrid,
                       double dMax,
                       double dMin,
                       double dInc,
                       int    iCycle)
  {
    this.stPlot.sXAxis      = new String(sText);
    this.stPlot.iXLog       = iGrid;
    this.stPlot.dXMaximum   = dMax;
    this.stPlot.dXMinimum   = dMin;
    this.stPlot.dXIncrement = dInc;
    this.stPlot.iXCycles    = iCycle;
    convertDataToPixels();
    repaint();
  }

  /** Method setYAxis()
   * <p> This method will reset the Y-Axis Parameters
   * @param sText = The Axis Label
   * @param iGrid = The type of grid to create
   * @param dMax  = The Maximum axis value
   * @param dMin  = The Minimum axis value
   * @param dInc  = The Increment value on the axis - Linear Grid only
   * @param iCycle = The Number of Cycles - Log Grid only
   */

  public void setYAxis(String sText,
                       int    iGrid,
                       double dMax,
                       double dMin,
                       double dInc,
                       int    iCycle)
  {
    this.stPlot.sYAxis      = new String(sText);
    this.stPlot.iYLog       = iGrid;
    this.stPlot.dYMaximum   = dMax;
    this.stPlot.dYMinimum   = dMin;
    this.stPlot.dYIncrement = dInc;
    this.stPlot.iYCycles    = iCycle;
    convertDataToPixels();
    repaint();
  }

  /** Method setData()
   *  <p> This method will reload the data.
   *  @param stPlot = Plot Data Structure.
   */

  public void setData(plotXYStruct st)
  {
	double dMax  = this.stPlot.dXMaximum;
	double dIncr = this.stPlot.dXIncrement;

    this.stPlot = st;

    iRows = stPlot.iRows;

    this.iLegend = stPlot.iLegend;
    if (stPlot.stLegend != null)
      this.stLegend = stPlot.stLegend;

    if (stPlot.stSymbols != null)
      this.stSymbol = stPlot.stSymbols;

    // Convert Raw Data to Pixels

    setAxes(dMax, dIncr);
//    convertDataToPixels();

    setRadius(dRadius);
    repaint();
  }

  /** Method setDepths()
   * <p> This method will set the depth ranges
   * @param dStart = starting depth
   * @param dEnd   = ending depth
   */

  public void setDepths(double dStart, double dEnd)
  {
    this.dStart    = dStart;
    this.dEnd      = dEnd;
    stPlot.sTitle3 = new String("" + dStart + "' to " + dEnd + "'");

    repaint();
  }

  /** Method setRadius()
   * <p> This method will set the borehole radius
   * @param dRadius = reference radius
   */

  public void setRadius(double dRadius)
  {
	int    iRatio = 0;
	double dXo    = 0.0;
	double dYo    = 0.0;
	double dRad   = 0.0;
	double dVal   = 0.0;

	this.dRadius = dRadius;

    iRatio = (int)((double)iWidth *
               (dXo - stPlot.dXMinimum) / (stPlot.dXMaximum - stPlot.dXMinimum));
    iXo = iXBegin + iRatio;

    iRatio = (int)((double)iHeight *
               (dYo - stPlot.dYMinimum) / (stPlot.dYMaximum - stPlot.dYMinimum));
    iYo = iHeight + iYBegin - iRatio;

    iXb = new int[iCircle];
    iYb = new int[iCircle];

    for (int i=0; i<iCircle; i++)
    {
	  dRad   = Math.PI * i / 180.0;

	  dVal   = dRadius * Math.sin(dRad);
	  iRatio = (int)((double)iWidth *
	             (dVal - stPlot.dXMinimum) / (stPlot.dXMaximum - stPlot.dXMinimum));
	  iXb[i] = iXBegin + iRatio;

	  dVal   = dRadius * Math.cos(dRad);
      iRatio = (int)((double)iHeight *
                 (dVal - stPlot.dYMinimum) / (stPlot.dYMaximum - stPlot.dYMinimum));
      iYb[i] = iHeight + iYBegin - iRatio;
	}

    repaint();
  }

  /* ================================================================== *
   * -------------------------- DRAW METHODS -------------------------- *
   * ================================================================== */

  /** Method drawTitles()
   *  <p> This method will draw the titles of the Plot.
   *  @param g - Graphics pointer.
   */

  public void drawTitles(Graphics g)
  {
    int iTemp = 0;

    g.setColor(Color.black);

    g.setFont( new Font("Monospaced", Font.BOLD, 14) );

    if (stPlot.sTitle1 != null)
    {
      iTemp = (int) (Math.abs(iWidth - (6 * stPlot.sTitle1.length())));
      g.drawString(stPlot.sTitle1, iTemp / 2, 10);
    }

    if (stPlot.sTitle2 != null)
    {
      iTemp = (int) (Math.abs(iWidth - (6 * stPlot.sTitle2.length())));
      g.drawString(stPlot.sTitle2, iTemp / 2, 30);
    }

    if (stPlot.sTitle3 != null)
    {
      iTemp = (int) (Math.abs(iWidth - (6 * stPlot.sTitle3.length())));
      g.drawString(stPlot.sTitle3, iTemp / 2, 50);
    }
  }

  /** Method drawLabels()
   *  <p> This method will draw the Linear Plot Labels.
   *  @param g - Graphics pointer.
   */

  public void drawLabels(Graphics g)
  {
    int ivalue;

    g.setColor(Color.black);

    if (stPlot.sXAxis != null)
    {
      ivalue = (iWidth+100)/2 - 3 * (stPlot.sXAxis).length();
      g.drawString(stPlot.sXAxis, ivalue, iYBegin+iHeight+35);
    }

    if (stPlot.sYAxis != null)
    {
      ivalue = (iHeight + 100) / 2 - 3 * (stPlot.sYAxis).length();

      Graphics2D g2 = (Graphics2D) g;
      g2.rotate( (Math.PI / 2.0));
      g2.drawString(stPlot.sYAxis, ivalue, -10);
      g2.rotate( -1.0 * (Math.PI / 2.0));
    }
  }

  /** Method drawGrid()
   *  <p> This method will draw the Grid Lines
   *  @param g - Graphics pointer.
   */

  public void drawGrid(Graphics g)
  {
    if (stPlot.iXLog == plot.plotXYStruct.LINEAR)
      drawLinearXAxisGrid(g);
    else
      drawLogXAxisGrid(g);

    if (stPlot.iYLog == plot.plotXYStruct.LINEAR)
      drawLinearYAxisGrid(g);
    else
      drawLogYAxisGrid(g);
  }

  /** Method drawLinearXAxisGrid()
   *  <p> This method will draw the Linear X-Axis Grid Lines
   *  @param g - Graphics pointer.
   */

  public void drawLinearXAxisGrid(Graphics g)
  {
    int    i      = 0;
    int    j      = 0;
    int    idraw  = 0;
    int    idel   = 0;
    int    iRatio = 0;
    int    iXinc  = 0;
    double dx     = 0.0;
    double del    = 0.0;

    iXinc = (int) ((double)iWidth * stPlot.dXIncrement /
                   (stPlot.dXMaximum - stPlot.dXMinimum));

    idraw = iXBegin;
    while (idraw < iXBegin+iWidth+5)
    {
      dx     = stPlot.dXMinimum + (double) i * stPlot.dXIncrement;
      iRatio = (int) ((double) iWidth *
                      (dx - stPlot.dXMinimum) /
                      (stPlot.dXMaximum - stPlot.dXMinimum));
      idraw = iXBegin + iRatio;

      if (idraw < iXBegin+iWidth+5)
      {
        g.setColor(Color.darkGray);
        g.drawLine(idraw, iYBegin, idraw, iYBegin+iHeight);

        g.setColor(Color.black);
        g.setFont( new Font("Monospaced", Font.BOLD, 10) );
        dx = Math.round(dx*1000)/1000.0;
        g.drawString(""+dx, idraw, iYBegin + iHeight + 15);

        for (j=1; j<5; j++)
        {
          idel = idraw + (j * iXinc) / 5;
          if (idel < iXBegin+iWidth)
          {
            g.setColor(Color.lightGray);
            g.drawLine(idel, iYBegin, idel, iYBegin+iHeight);
          }
        }
      }

      i++;
    }
  }

  /** Method drawLogXAxisGrid()
   *  <p> This method will draw the Log X-Axis Grid Lines
   *  @param g - Graphics pointer.
   */

  public void drawLogXAxisGrid(Graphics g)
  {
    int    i, j;
    int    l=1;
    int    k=0;
    int    iXinc;
    double dLabel = 0.0;
    double del    = 0.0;
    int    jXinc;
    double dTemp;
    double dXinc;

    iXinc = 1 + iWidth / stPlot.iXCycles;

    g.setColor(Color.darkGray);
    g.drawLine(iXBegin, iYBegin, iXBegin, iYBegin+iHeight);

    for (i=iXBegin+iXinc; i<iXBegin+iWidth+10; i+=iXinc)
    {
      g.setColor(Color.darkGray);
      g.drawLine(i, iYBegin, i, iYBegin+iHeight);

      g.setColor(Color.lightGray);
      for (j=2; j<10; j++)
      {
        dXinc = (double) j;
        dTemp = 1.0+(double) iXinc * Math.log(dXinc) / Math.log(10.0);
        jXinc = i - iXinc + (int) dTemp;

        g.drawLine(jXinc, iYBegin, jXinc, iYBegin+iHeight);
      }
    }

    g.setFont( new Font("Monospaced", Font.BOLD, 10) );
    g.setColor(Color.black);
    k=1;
    del = stPlot.dXIncrement;
    g.drawString("" + stPlot.dXMinimum, iXBegin-10, iYBegin+iHeight+15);
    for (i=iXBegin+iXinc; i<iXBegin+iWidth+10; i+=iXinc)
    {
      g.setColor(Color.darkGray);

      dLabel = (double) k * 10.0 * stPlot.dXMinimum;
      k = k * 10;

      g.setColor(Color.black);
      g.drawString("" + dLabel, i-10, iYBegin+iHeight+15);
      del = stPlot.dXIncrement + del;
    }
  }

  /** Method drawLinearYAxisGrid()
   *  <p> This method will draw the Linear Y-Axis Grid Lines
   *  @param g - Graphics pointer.
   */

  public void drawLinearYAxisGrid(Graphics g)
  {
    int    i = 0;
    int    j = 0;
    int    idraw  = 0;
    int    idel   = 0;
    int    iRatio = 0;
    int    iYinc  = 0;
    double dy     = 0.0;
    double del    = 0.0;

    iYinc = (int) ((double)iHeight * stPlot.dYIncrement /
                   (stPlot.dYMaximum - stPlot.dYMinimum));

    idraw = iYBegin + iHeight;
    while (idraw > iYBegin-5)
    {
      dy = stPlot.dYMinimum + (double) i * stPlot.dYIncrement;
      iRatio = (int)((double)iHeight *
                     (dy - stPlot.dYMinimum) /
                     (stPlot.dYMaximum - stPlot.dYMinimum));
      idraw = iHeight + iYBegin - iRatio;

      if (idraw > iYBegin-5)
      {
        g.setColor(Color.darkGray);
        g.drawLine(iXBegin, idraw, iXBegin+iWidth, idraw);

        g.setColor(Color.black);
        g.setFont( new Font("Monospaced", Font.BOLD, 10) );
        dy = Math.round(dy*1000)/1000.0;
        g.drawString("" + dy, iXBegin-45, idraw);

        for (j=1; j<5; j++)
        {
          idel = idraw - (j * iYinc) / 5;
          if (idel > iYBegin)
          {
            g.setColor(Color.lightGray);
            g.drawLine(iXBegin, idel, iXBegin+iWidth, idel);
          }
        }
      }

      i++;
    }
  }

  /** Method drawLogYAxisGrid()
   *  <p> This method will draw the Log X-Axis Grid Lines
   *  @param g - Graphics pointer.
   */

  public void drawLogYAxisGrid(Graphics g)
  {
    int    i = 0;
    int    j = 0;
    int    k = 1;
    int    iYinc  = 0;
    int    jXinc  = 0;
    double dy     = 0.0;
    double dTemp  = 0.0;
    double dXinc  = 0.0;

    iYinc = 1 + iHeight / stPlot.iYCycles;

    for (i=iYBegin+iHeight; i>iYBegin; i-=iYinc)
    {
      k = i;
      g.setColor(Color.darkGray);
      g.drawLine(iXBegin, i, iXBegin+iWidth, i);

      g.setColor(Color.lightGray);
      for (j=2; j<10; j++)
      {
        dXinc = (double) j;
        dTemp = 1.0+(double) iYinc * Math.log(dXinc) / Math.log(10.0);
        jXinc = i - (int) dTemp;

        g.drawLine(iXBegin, jXinc, iXBegin+iWidth, jXinc);
      }
    }

    g.setColor(Color.darkGray);
    g.drawLine(iXBegin, k-iYinc, iXBegin+iWidth, k-iYinc);

    g.setFont( new Font("Monospaced", Font.BOLD, 10) );
    g.setColor(Color.black);
    k=1;
    g.drawString("" + stPlot.dYMinimum, iXBegin-45, iYBegin+iHeight);
    for (i=iYBegin+iHeight; i>iYBegin; i-=iYinc)
    {
      dy = (double) k * 10.0 * stPlot.dYMinimum;
      k = k * 10;
      g.drawString("" + dy, iXBegin-45, i-iYinc);
    }
  }

  /** Method drawPlotData()
   * <p> This method will plot the data
   *  @param g - Graphics pointer.
   */

  public void drawPlotData(Graphics g)
  {
    int iDraw = 0;
    int iX1   = 0;
    int iX2   = 0;
    int iY1   = 0;
    int iY2   = 0;

    for (int i=0; i<iRows; i++)
    {
      iX1 = iX[i];
      iY1 = iY[i];

      if ( (iX1 > iXBegin) && (iX1 < iXBegin + iWidth)  &&
           (iY1 > iYBegin) && (iY1 < iYBegin + iHeight))
      {
        iDraw = 0;
        if ((stPlot.depths[i] >= dStart) && (stPlot.depths[i] < dEnd))
          iDraw = 1;

        if (iDraw == 1)
        {
          pSymbol.drawSymbol(g, stSymbol[i], iX[i], iY[i]);
        }
      }
    }
  }

  /** Method drawLegend()
   *  <p> This method will create the Legend
   *  @param g        = Graphics pointer.
   */

  public void drawLegend(Graphics g)
  {
    g.setFont( new Font("Monospaced", Font.BOLD, 11) );
    g.setColor(Color.black);

    if (stPlot.sampling.length() > 0)
      g.drawString(stPlot.sampling, iXBegin+iWidth+10, iYBegin-26);

    if (stPlot.sLegTitle.length() > 0)
      g.drawString(stPlot.sLegTitle, iXBegin+iWidth+10, iYBegin-13);

    for (int i=0; i<iLegend; i++)
    {
      pSymbol.drawSymbol(g, stLegend[i], iXBegin+iWidth+15, iYBegin+i*13);

      g.setColor(Color.black);
      g.drawString(stLegend[i].symbol, iXBegin+iWidth+15+8, iYBegin+4+i*13);
    }
  }

  /** Method draw()
   *  <p> This method will create the Linear Plot.
   *  @param g - Graphics pointer.
   */

  public void draw(Graphics g)
  {
	int    len = 0;
	String txt = "";

    drawTitles(g);
    drawLabels(g);
    drawGrid(g);

    drawPlotData(g);
    drawLegend(g);

    g.setColor(Color.black);
    g.setFont( new Font("Monospaced", Font.BOLD, 12) );
    g.drawPolygon(iXb, iYb, iCircle);
    txt = "Radius = "+dRadius + "\"";
    len = 4*txt.length();
    g.drawString(txt, iXo-len, iYo);
  }

  /** Method paint()
   *  <p> This method will create the Linear Plot.
   *  @param g - Graphics pointer.
   */

  public void paint(Graphics g)
  {
	draw(g);
  }
}

/*
 *  @version 1.1 4/25/2017
 *  @author  John Victorine
 */
